// CS205 Fall 2006
// Grid.java

import java.util.Random;

public class Grid {
	// OVERVIEW: A Grid is a 2-dimensional grid.  Each square in the grid is either
	//    empty or contains a SimObject object.
	//
	//    A typical Grid is: [ [ s_0_ymax - 1, s_1_ymax - 1, ..., s_xmax - 1_ymax - 1 ],
	//						   ... ,
	//                         [ s_0_1, s_1_1, ..., s_xmax - 1_1],
	//                         [ s_0_0, s_1_0, ..., s_xmax - 1_0] ]
	//						 
	//    Where each s_x_y is either a SimObject object or empty.
	//

	protected int x_max; //invariant x_max > 0;
	protected int y_max; //invariant y_max > 0;

	/*nonnull*/SimObject[][] simobjects;
	/*nonnull*/private Random rand = new Random();

	// invariant for each s: SimObject in simobjects
	//     s.isInitialized
	//     s.x () >= 0 && s.x () <= x_max
	//     s.y () >= 0 && s.y () <= y_max
	//     no other t: SimObject in simobjects with the same x and y

	Grid(int p_x_max, int p_y_max)
	// REQUIRES: p_x_max and p_y_max must be > 0
	// EFFECTS: Constructs a new grid with the specified maximum x and y
	// 		   values.
	{
		x_max = p_x_max;
		y_max = p_y_max;
		simobjects = new SimObject[x_max + 1][y_max + 1];
	}

	public int maxX()
	// EFFECTS: Returns the maximum X value for this grid.
	{
		return x_max;
	}

	public int maxY()
	// EFFECTS: Returns the maximum Y value for this grid.
	{
		return y_max;
	}

	synchronized public void setObjectAt(int x, int y, /*nonnull*/
			SimObject newObject)
	// REQUIRES: newObject is not initialized
	// EFFECTS: places newObject on the grid at location (x,y) (replacing any
	//    object currently on the grid at that location).
	{
		// If an object already exists in the grid with the specified
		// row and column, remove it before adding the new one.

		newObject.init(x, y, this);
		simobjects[x][y] = newObject;
	}

	public Coordinate randomFreeLocation()
	// REQUIRES: there is at least one free square on the grid
	// EFFECTS: returns a new location on the grid that is unoccupied
	//    Note: client should lock the grid before calling this, and
	//       hold the lock until the square is used, otherwise the square
	//       may be occupied before the object is placed in it.
	{
		// Dangerous implementation...if there is no free square, this will 
		// loop forever... (but requires was not satisfied)

		while (true) {
			int xloc = rand.nextInt(x_max);
			int yloc = rand.nextInt(y_max);

			if (getObjectAt(xloc, yloc) == null) { return new Coordinate(xloc,
					yloc); }
		}
	} 

	synchronized public void setObjectAt(/*nonnull*/Coordinate coord, /*nonnull*/
			SimObject newObject)
	// REQUIRES: newObject is not initialized
	// EFFECTS: places newObject on the grid at location (x,y) (replacing any
	//    object currently on the grid at that location).
	{
		setObjectAt(coord.getX(), coord.getY(), newObject);
	}

	synchronized public void removeAllObjects()
	// EFFECTS: removes all objects from the grid.
	{
		for (int x = 0; x <= maxX(); x++) {
			for (int y = 0; y <= maxY(); y++) {
				simobjects[x][y] = null;
			}
		}
	}

	synchronized public void removeObjectAt(int x, int y)
	// EFFECTS: If there is an object at (x, y), removes it.
	{
		simobjects[x][y] = null;
	}

	synchronized public void removeObjectAt(/*nonnull*/Coordinate c)
	// EFFECTS: If there is an object at c, removes it.
	{
		removeObjectAt(c.getX(), c.getY());
	}

	synchronized public boolean isSquareEmpty(int x, int y)
			throws RuntimeException
	// REQUIRES: 0 >= x < columns && 0 >= y < rows
	// EFFECTS: Returns the Cell at the specified x, y location.
	//          Returns null if nothing is present there.
	{
		return (getObjectAt(x, y) == null);
	}

	public boolean validLocation(/*nonnull*/Coordinate c) {
		return validLocation(c.getX(), c.getY());
	}

	public boolean validLocation(int x, int y)
	// EFFECTS: Determines if x, y is a valid grid location.
	{
		return (x >= 0 && x <= maxX()
					&& y >= 0 && y <= maxY ());
	}

	// We need the unsynchronized version of getObjectAt for the display --- we want to 
	// be able to display the grid without acquiring a lock.  This runs the risk that
	// the grid will be temporarily displayed in an inconsistent state.

	/* unsynchronized*/public SimObject grabObjectAt(int x, int y)
			throws RuntimeException
	// REQUIRES: 0 >= row < rows && 0 >= col < columns
	// EFFECTS: Returns the Cell at the specified row and column.
	//          Returns null if nothing is present there.
	{
		if ((x < 0) || (x > maxX())) { throw new RuntimeException(
				"Bad x parameter to getObjectAt: " + x); }

		if ((y < 0) || (y > maxY())) { throw new RuntimeException(
				"Bad y parameter to getObjectAt: " + y); }

		return simobjects[x][y];
	}

	synchronized public SimObject getObjectAt(int x, int y)
			throws RuntimeException
	// REQUIRES: 0 >= x < columns && 0 >= y < rows
	// EFFECTS: Returns the Cell at the specified location
	//          Returns null if nothing is present there.
	{
		return grabObjectAt(x, y);
	}

	synchronized public SimObject getObjectAt(/*nonnull*/Coordinate c)
			throws RuntimeException
	// REQUIRES: 0 >= x < columns && 0 >= y < rows
	// EFFECTS: Returns the Cell at the specified location
	//          Returns null if nothing is present there.
	{
		return grabObjectAt(c.getX(), c.getY());
	}
}
